import numpy

def add_customizations(estimator, customizations):
	for customization in customizations:
		if not isinstance(customization, Customization):
			raise TypeError()
	if hasattr(estimator, "pmml_customizations_"):
		pmml_customizations = estimator.pmml_customizations_
		pmml_customizations = numpy.append(pmml_customizations, customizations)
	else:
		pmml_customizations = numpy.asarray(customizations)
	estimator.pmml_customizations_ = pmml_customizations

def clear_customizations(estimator):
	if hasattr(estimator, "pmml_customizations_"):
		delattr(estimator, "pmml_customizations_")

class Customization(object):

	def __init__(self, command, xpath_expr, pmml_element):
		commands = ["insert", "update", "delete"]
		if command not in commands:
			raise ValueError("Command {} not in {}".format(command, commands))
		self.command = command
		if xpath_expr is not None:
			if not isinstance(xpath_expr, str):
				raise TypeError("XPath expression is not a string")
		else:
			if command in ["delete"]:
				raise ValueError("Command {} requires XPath expression".format(command))
		self.xpath_expr = xpath_expr
		# XXX: isinstance(pmml, sklearn2pmml.metrics.PMMLElement)
		if hasattr(pmml_element, "tostring"):
			pmml_element = pmml_element.tostring()
		if pmml_element is not None:
			if not isinstance(pmml_element, str):
				raise TypeError("PMML element is not a string")
			if command in ["delete"]:
				raise ValueError("Command {} does not support PMML element".format(command))
		else:
			if command in ["insert", "update"]:
				raise ValueError("Command {} requires PMML element".format(command))
		self.pmml_element = pmml_element



from sklearn.base import clone, BaseEstimator, TransformerMixin
from sklearn.neural_network import MLPRegressor

class MLPTransformer(BaseEstimator, TransformerMixin):

	def __init__(self, mlp, transformer_output_layer = -1):
		if not isinstance(mlp, MLPRegressor):
			raise TypeError("The mlp object is not an instance of {0}".format(MLPRegressor.__name__))
		self.mlp = mlp
		self.transformer_output_layer = transformer_output_layer

	def fit(self, X, y = None):
		self.mlp_ = clone(self.mlp)
		self.mlp_.fit(X, X)
		if self.transformer_output_layer != -1:
			self.mlp_.n_layers_ = self.transformer_output_layer + 1
			self.mlp_.n_outputs_ = self.mlp_.hidden_layer_sizes[self.transformer_output_layer - 1]
		return self

	def transform(self, X):
		return self.mlp_.predict(X)


from sklearn.metrics import accuracy_score, confusion_matrix, fbeta_score, mean_absolute_error, mean_squared_error, precision_score, r2_score, recall_score
from sklearn2pmml.util.pmml import make_element
from sklearn2pmml.util.pmml import Array, PMMLElement

import numpy

class ConfusionMatrix(PMMLElement):

	def __init__(self, labels, confusion_matrix):
		super(ConfusionMatrix, self).__init__(make_element("ConfusionMatrix"))

		pmml_class_labels = make_element("ClassLabels")
		pmml_class_labels.append(Array(labels).root)
		self.append(pmml_class_labels)

		pmml_matrix = make_element("Matrix")
		for row in confusion_matrix:
			pmml_matrix.append(Array(row).root)
		self.append(pmml_matrix)

class ModelExplanation(PMMLElement):

	def __init__(self):
		super(ModelExplanation, self).__init__(make_element("ModelExplanation"))

class PredictiveModelQuality(PMMLElement):

	def __init__(self, estimator, X, y, target_field, data_usage, data_name):
		super(PredictiveModelQuality, self).__init__(make_element("PredictiveModelQuality", targetField = target_field))

		self.estimator = estimator

		self.X = X
		self.y = y

		# XXX
		self.yt = estimator.predict(X)

		if data_usage is not None:
			data_usages = ["training", "test", "validation"]
			if data_usage not in data_usages:
				raise ValueError("Data usage {0} not in {1}".format(data_usage, data_usages))
			self.set("dataUsage", data_usage)
		if data_name is not None:
			self.set("dataName", data_name)

		n_samples = X.shape[0]
		n_features = (estimator.n_features_in_ if hasattr(estimator, "n_features_in_") else X.shape[1])

		self.set("numOfRecords", n_samples)
		self.set("numOfPredictors", n_features)

class ClassifierQuality(PredictiveModelQuality):

	def __init__(self, estimator, X, y, target_field, data_usage = "training", data_name = None):
		super(ClassifierQuality, self).__init__(estimator = estimator, X = X, y = y, target_field = target_field, data_usage = data_usage, data_name = data_name)

	@property
	def classes_(self):
		return self.estimator.classes_

	def with_all_metrics(self):
		return self \
			.with_accuracy() \
			.with_confusion_matrix()

	def with_accuracy(self):
		accuracy = accuracy_score(self.y, self.yt)
		return self.set("accuracy", accuracy)

	def with_confusion_matrix(self):
		pmml_confusion_matrix = ConfusionMatrix(labels = self.classes_, confusion_matrix = confusion_matrix(self.y, self.yt, normalize = None))
		return self.append(pmml_confusion_matrix)

class BinaryClassifierQuality(ClassifierQuality):

	def __init__(self, estimator, X, y, target_field, data_usage = "training", data_name = None):
		super(BinaryClassifierQuality, self).__init__(estimator = estimator, X = X, y = y, target_field = target_field, data_usage = data_usage, data_name = data_name)

		if len(self.classes_) != 2:
			raise ValueError()

	def with_all_metrics(self):
		return super(BinaryClassifierQuality, self).with_all_metrics() \
			.with_precision() \
			.with_recall() \
			.with_F1() \
			.with_F2() \
			.with_Fhalf()

	def with_precision(self):
		classes = self.classes_
		precision = precision_score(self.y, self.yt, labels = classes, pos_label = classes[1])
		return self.set("precision", precision)

	def with_recall(self):
		classes = self.classes_
		recall = recall_score(self.y, self.yt, labels = classes, pos_label = classes[1])
		return self.set("recall", recall)

	def with_F1(self):
		return self.set("F1", self._fbeta_score(1.0))

	def with_F2(self):
		return self.set("F2", self._fbeta_score(2.0))

	def with_Fhalf(self):
		return self.set("Fhalf", self._fbeta_score(0.5))

	def _fbeta_score(self, beta):
		classes = self.classes_
		return fbeta_score(self.y, self.yt, beta = beta, labels = classes, pos_label = classes[1])

class RegressorQuality(PredictiveModelQuality):
	
	def __init__(self, estimator, X, y, target_field, data_usage = "training", data_name = None):
		super(RegressorQuality, self).__init__(estimator = estimator, X = X, y = y, target_field = target_field, data_usage = data_usage, data_name = data_name)

	def with_all_metrics(self):
		return self \
			.with_ME() \
			.with_MAE() \
			.with_MSE() \
			.with_RMSE() \
			.with_R2() \
			.with_SSE() \
			.with_SSR()

	def with_ME(self):
		me = numpy.sum(self.yt - self.y) / self.X.shape[0]
		return self.set("meanError", me)

	def with_MAE(self):
		mae = mean_absolute_error(self.y, self.yt)
		return self.set("meanAbsoluteError", mae)

	def with_MSE(self):
		mse = mean_squared_error(self.y, self.yt)
		return self.set("meanSquaredError", mse)

	def with_RMSE(self):
		rmse = numpy.sqrt(mean_squared_error(self.y, self.yt))
		return self.set("rootMeanSquaredError", rmse)

	def with_R2(self):
		r2 = r2_score(self.y, self.yt)
		return self.set("r-squared", r2)

	def with_SSE(self):
		sse = numpy.sum((self.y - self.yt) ** 2)
		return self.set("sumSquaredError", sse)

	def with_SSR(self):
		ymean = numpy.mean(self.y)
		ssr = numpy.sum((ymean - self.yt) ** 2)
		return self.set("sumSquaredRegression", ssr)



from sklearn.base import clone, BaseEstimator, ClassifierMixin, RegressorMixin
try:
	from sklearn.linear_model import LinearRegression
	from sklearn.linear_model._base import LinearClassifierMixin, LinearModel, SparseCoefMixin
except ImportError:
	from sklearn.linear_model.base import LinearClassifierMixin, LinearModel, LinearRegression, SparseCoefMixin
from sklearn.preprocessing import OneHotEncoder
from sklearn.utils.metaestimators import _BaseComposition
from sklearn2pmml import _is_ordinal
from sklearn2pmml.util import eval_rows, fqn, to_expr_func, Predicate

import copy
import numpy

def _checkGBDTRegressor(gbdt):
	if hasattr(gbdt, "apply"):
		return gbdt
	else:
		try:
			from lightgbm import LGBMRegressor
			if isinstance(gbdt, LGBMRegressor):
				return gbdt
		except ImportError:
			pass
	raise TypeError("GBDT object (class {0}) is not supported".format(fqn(gbdt)))

def _checkLM(lm):
	if isinstance(lm, (LinearModel, LinearRegression, SparseCoefMixin)):
		return lm
	raise TypeError("LM object (class {0}) is not supported".format(fqn(lm)))

def _checkGBDTClassifier(gbdt):
	if hasattr(gbdt, "apply"):
		return gbdt
	else:
		try:
			from lightgbm import LGBMClassifier
			if isinstance(gbdt, LGBMClassifier):
				return gbdt
		except ImportError:
			pass
	raise TypeError("GBDT object (class {0}) is not supported".format(fqn(gbdt)))

def _checkLR(lr):
	if isinstance(lr, LinearClassifierMixin):
		return lr
	raise TypeError("LR object (class {0}) is not supported".format(fqn(lr)))

def _extract_step_params(name, params):
	prefix = name + "__"
	step_params = dict()
	for k, v in (params.copy()).items():
		if k.startswith(prefix):
			step_params[k[len(prefix):len(k)]] = v
			del params[k]
	return step_params

def _mask_params(params, mask):
	masked_params = dict()
	for k, v in params.items():
		if isinstance(v, numpy.ndarray):
			if v.shape[0] != mask.shape[0]:
				raise ValueError()
			v = v[mask]
		masked_params[k] = v
	return masked_params

class GBDTEstimator(BaseEstimator):

	def __init__(self, gbdt):
		self.gbdt = gbdt

	def _leaf_indices(self, X):
		# Scikit-Learn
		# XGBoost
		if hasattr(self.gbdt_, "apply"):
			id = self.gbdt_.apply(X)
			if id.ndim > 2:
				id = id[:, :, 0]
		# LightGBM
		else:
			id = self.gbdt_.predict(X, pred_leaf = True)
		return id

	def _encoded_leaf_indices(self, X):
		id = self._leaf_indices(X)
		idt = self.ohe_.transform(id)
		return idt

class GBDTLMRegressor(GBDTEstimator, RegressorMixin):

	def __init__(self, gbdt, lm):
		super(GBDTLMRegressor, self).__init__(_checkGBDTRegressor(gbdt))
		self.lm = _checkLM(lm)

	def fit(self, X, y, **fit_params):
		self.gbdt_ = clone(self.gbdt)
		self.gbdt_.fit(X, y, **_extract_step_params("gbdt", fit_params))
		id = self._leaf_indices(X)
		self.ohe_ = OneHotEncoder(categories = "auto")
		self.ohe_.fit(id)
		idt = self.ohe_.transform(id)
		self.lm_ = clone(self.lm)
		self.lm_.fit(idt, y, **_extract_step_params("lm", fit_params))
		return self

	def predict(self, X):
		idt = self._encoded_leaf_indices(X)
		return self.lm_.predict(idt)

class GBDTLRClassifier(GBDTEstimator, ClassifierMixin):

	def __init__(self, gbdt, lr):
		super(GBDTLRClassifier, self).__init__(_checkGBDTClassifier(gbdt))
		self.lr = _checkLR(lr)

	@property
	def classes_(self):
		return self.lr_.classes_

	def fit(self, X, y, **fit_params):
		self.gbdt_ = clone(self.gbdt)
		self.gbdt_.fit(X, y, **_extract_step_params("gbdt", fit_params))
		id = self._leaf_indices(X)
		self.ohe_ = OneHotEncoder(categories = "auto")
		self.ohe_.fit(id)
		idt = self.ohe_.transform(id)
		self.lr_ = clone(self.lr)
		self.lr_.fit(idt, y, **_extract_step_params("lr", fit_params))
		return self

	def predict(self, X):
		idt = self._encoded_leaf_indices(X)
		return self.lr_.predict(idt)

	def predict_proba(self, X):
		idt = self._encoded_leaf_indices(X)
		return self.lr_.predict_proba(idt)

def _codes(y):
	if hasattr(y, "cat"):
		return (y.cat).codes.values
	else:
		return y.codes

class OrdinalClassifier(BaseEstimator, ClassifierMixin):

	def __init__(self, estimator):
		self.estimator = estimator

	def fit(self, X, y, **fit_params):
		dtype = y.dtype
		if not _is_ordinal(dtype):
			raise TypeError()
		y_codes = _codes(y)
		self.classes_ = numpy.asarray(dtype.categories)
		self.estimators_ = []
		n_estimators = len(self.classes_) - 1
		for i in range(n_estimators):
			y_bin = (y_codes > i).astype(int)
			estimator = clone(self.estimator)
			estimator.fit(X, y_bin, **fit_params)
			self.estimators_.append(estimator)
		return self

	def predict(self, X, **predict_params):
		proba = self.predict_proba(X, **predict_params)
		indices = numpy.argmax(proba, axis = 1)
		return numpy.take(self.classes_, indices)

	def predict_proba(self, X, **predict_proba_params):
		n_estimators = len(self.classes_) - 1
		proba = []
		for i in range(n_estimators):
			proba.append(self.estimators_[i].predict_proba(X, **predict_proba_params)[:, 1])
		result = []
		for i in range(n_estimators + 1):
			if i == 0:
				result.append(1 - proba[0])
			elif i == n_estimators:
				result.append(proba[-1])
			else:
				result.append(proba[i - 1] - proba[i])
		return numpy.asarray(result).T

def _to_sparse(X, step_mask, step_result):
	# Make array
	if len(step_result.shape) == 1:
		result = numpy.empty((X.shape[0], ), dtype = object)
	else:
		result = numpy.empty((X.shape[0], step_result.shape[1]), dtype = object)
	# Fill array
	if len(step_result.shape) == 1:
		result[step_mask.ravel()] = step_result
	else:
		result[step_mask.ravel(), :] = step_result
	return result

class _BaseEnsemble(_BaseComposition):

	def __init__(self, steps, controller):
		for step in steps:
			if type(step) is not tuple:
				raise TypeError("Step is not a tuple")
			if len(step) != 3:
				raise TypeError("Step is not a three-element (name, estimator, predicate) tuple")
			name, estimator, predicate = step
			if not isinstance(predicate, (str, Predicate)):
				raise TypeError()
		self.steps = steps
		if controller:
			if not hasattr(controller, "transform"):
				raise TypeError()
		self.controller = controller

	@property
	def _steps(self):
		return [(name, estimator) for name, estimator, predicate in self.steps]

	@_steps.setter
	def _steps(self, value):
		self.steps = [(name, estimator, predicate) for ((name, estimator), (_, _, predicate)) in zip(value, self.steps)]

	def get_params(self, deep = True):
		return self._get_params("_steps", deep = deep)

	def set_params(self, **kwargs):
		self._set_params("_steps", **kwargs)
		return self

	def _to_evaluation_dataset(self, X):
		if self.controller is not None:
			return self.controller.transform(X)
		return X

class Link(BaseEstimator):

	def __init__(self, estimator, augment_funcs, prefit = False):
		self.estimator = estimator
		self.augment_funcs = augment_funcs
		self.prefit = prefit
		if prefit:
			self.estimator_ = copy.deepcopy(self.estimator)

	def fit(self, X, y, **fit_params):
		self.estimator_ = clone(self.estimator)
		self.estimator_.fit(X, y, **fit_params)
		return self

	def predict(self, X):
		return self.estimator_.predict(X)

	def predict_proba(self, X):
		return self.estimator_.predict_proba(X)

	def augment(self, X):
		Y = None
		for augment_func in self.augment_funcs:
			yt = getattr(self.estimator_, augment_func)(X)
			yt = yt.reshape(X.shape[0], -1)
			if Y is None:
				Y = yt
			else:
				Y = numpy.hstack((Y, yt))
		if Y is None:
			return X
		else:
			return numpy.hstack((X, Y))

class EstimatorChain(_BaseEnsemble):

	def __init__(self, steps, controller = None, multioutput = True):
		super(EstimatorChain, self).__init__(steps, controller)
		self.multioutput = multioutput

	def fit(self, X, y, **fit_params):
		if len(y.shape) > 1:
			if len(self.steps) != y.shape[1]:
				raise ValueError()
			y = numpy.asarray(y)
		X_eval = self._to_evaluation_dataset(X)
		i = 0
		for name, estimator, predicate in self.steps:
			step_mask_func = to_expr_func(predicate)
			step_mask = eval_rows(X_eval, step_mask_func, dtype = bool)
			if numpy.sum(step_mask) < 1:
				raise ValueError(predicate)
			step_X = X[step_mask]
			if len(y.shape) == 1:
				step_y = y[step_mask]
			elif len(y.shape) == 2:
				step_y = y[step_mask, i]
			else:
				raise ValueError()
			step_fit_params = _mask_params(_extract_step_params(name, fit_params), step_mask)
			estimator.fit(step_X, step_y, **step_fit_params)
			if isinstance(estimator, Link):
				X = estimator.augment(X)
			i += 1
		return self

	def _predict(self, X, predict_method):
		result = None
		X_eval = self._to_evaluation_dataset(X)
		for name, estimator, predicate in self.steps:
			step_mask_func = to_expr_func(predicate)
			step_mask = eval_rows(X_eval, step_mask_func, dtype = bool)
			if numpy.sum(step_mask) < 1:
				continue
			step_X = X[step_mask]
			step_result = getattr(estimator, predict_method)(step_X)
			step_result = _to_sparse(X, step_mask, step_result)
			if self.multioutput:
				step_result = step_result.reshape(X.shape[0], -1)
			if result is None:
				result = step_result
			else:
				if self.multioutput:
					result = numpy.hstack((result, step_result))
				else:
					if step_result.shape != result.shape:
						raise ValueError()
					result[step_mask] = step_result[step_mask]
			if isinstance(estimator, Link):
				X = estimator.augment(X)
		return result

	def predict(self, X):
		return self._predict(X, "predict")

	def predict_proba(self, X):
		return self._predict(X, "predict_proba")

class SelectFirstEstimator(_BaseEnsemble):

	def __init__(self, steps, controller, eval_rows):
		super(SelectFirstEstimator, self).__init__(steps, controller)
		self.eval_rows = eval_rows

	def _eval_step_mask(self, X, predicate):
		step_func = to_expr_func(predicate)
		if self.eval_rows:
			return eval_rows(X, step_func, dtype = bool)
		else:
			return step_func(X)

	def fit(self, X, y, **fit_params):
		X_eval = self._to_evaluation_dataset(X)
		mask = numpy.full(X.shape[0], fill_value = False)
		for name, estimator, predicate in self.steps:
			step_mask = self._eval_step_mask(X_eval, predicate)
			step_mask[mask] = False
			if numpy.sum(step_mask) < 1:
				raise ValueError(predicate)
			step_X = X[step_mask]
			step_y = y[step_mask]
			step_fit_params = _mask_params(_extract_step_params(name, fit_params), step_mask)
			estimator.fit(step_X, step_y, **step_fit_params)
			mask = numpy.logical_or(mask, step_mask)
		return self

	def _predict(self, X, predict_method):
		result = None
		X_eval = self._to_evaluation_dataset(X)
		mask = numpy.full(X.shape[0], fill_value = False)
		for name, estimator, predicate in self.steps:
			step_mask = self._eval_step_mask(X_eval, predicate)
			step_mask[mask] = False
			if numpy.sum(step_mask) < 1:
				continue
			step_X = X[step_mask]
			step_result = getattr(estimator, predict_method)(step_X)
			step_result = _to_sparse(X, step_mask, step_result)
			if result is None:
				result = step_result
			else:
				result[step_mask] = step_result[step_mask]
			mask = numpy.logical_or(mask, step_mask)
		return result

	def apply(self, X):
		return self._predict(X, "apply")

	def predict(self, X):
		return self._predict(X, "predict")

class SelectFirstRegressor(SelectFirstEstimator, RegressorMixin):

	def __init__(self, steps, controller = None, eval_rows = True):
		super(SelectFirstRegressor, self).__init__(steps, controller, eval_rows)

class SelectFirstClassifier(SelectFirstEstimator, ClassifierMixin):

	def __init__(self, steps, controller = None, eval_rows = True):
		super(SelectFirstClassifier, self).__init__(steps, controller, eval_rows)

	def predict_proba(self, X):
		return self._predict(X, "predict_proba")



from sklearn.base import clone, BaseEstimator, TransformerMixin
from sklearn2pmml.preprocessing import ExpressionTransformer

import copy

class BusinessDecisionTransformer(BaseEstimator, TransformerMixin):

	def __init__(self, transformer, business_problem, decisions, prefit = True):
		if transformer is None:
			pass
		elif isinstance(transformer, str):
			pass
		elif isinstance(transformer, TransformerMixin):
			pass
		else:
			raise TypeError("The transformer object is not an instance of {0}".format(TransformerMixin.__name__))
		self.transformer = transformer
		self.business_problem = business_problem
		for decision in decisions:
			if type(decision) is not tuple:
				raise TypeError("Decision is not a tuple")
			if len(decision) != 2:
				raise TypeError("Decision is not a two-element (value, description) tuple")
		self.decisions = decisions
		self.prefit = prefit
		if prefit:
			self.transformer_ = self._make_transformer(prefit = True)

	def _make_transformer(self, prefit = True):
		if self.transformer is None:
			return None
		elif isinstance(self.transformer, str):
			return ExpressionTransformer(self.transformer)
		elif isinstance(self.transformer, TransformerMixin):
			if prefit:
				return copy.deepcopy(self.transformer)
			return clone(self.transformer)
		else:
			raise TypeError("The transformer object is not an instance of {0}".format(TransformerMixin.__name__))

	def fit(self, X, y = None):
		self.transformer_ = self._make_transformer(prefit = False)
		if self.transformer_ is not None:
			if y is None:
				self.transformer_.fit(X)
			else:
				self.transformer_.fit(X, y)
		return self

	def transform(self, X):
		if self.transformer_ is not None:
			return self.transformer_.transform(X)
		return X



from pycaret.internal.preprocess.transformers import TransformerWrapper
from sklearn2pmml import _escape as _default_escape
from sklearn2pmml import make_pmml_pipeline as default_make_pmml_pipeline

def _escape(obj, escape_func):
	obj = _default_escape(obj, escape_func = escape_func)
	if isinstance(obj, TransformerWrapper):
		obj.transformer = escape_func(obj.transformer, escape_func = escape_func)
	return obj

def make_pmml_pipeline(obj, active_fields = None, target_fields = None, escape_func = _escape):
	return default_make_pmml_pipeline(obj, active_fields = active_fields, target_fields = target_fields, escape_func = escape_func)



try:
	from CHAID import Tree
except ImportError:
	pass
from pandas import DataFrame, Series
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin

import numpy
import pandas
import warnings

class CHAIDEstimator(BaseEstimator):

	def __init__(self, config):
		self.config = config

	def _target_type(self):
		raise NotImplementedError()

	def fit(self, X, y, **fit_params):
		if isinstance(X, DataFrame):
			features = dict()
			for col in X.columns.values:
				features[col] = "nominal"
			df = pandas.concat((X, y), axis = 1)
			tree = Tree.from_pandas_df(df, i_variables = features, d_variable = y.name, dep_variable_type = self._target_type(), **self.config)
		else:
			variable_types = ["nominal"] * X.shape[1]
			tree = Tree.from_numpy(X, y, variable_types = variable_types, dep_variable_type = self._target_type(), **self.config)
		treelib_tree = tree.to_tree()
		self.tree_ = tree
		self.treelib_tree_ = treelib_tree
		return self

	def apply(self, X):
		if X is not None:
			warnings.warn("Ignoring the 'X' argument. Returning leaf indices for the training dataset")
		return self.tree_.node_predictions()

class CHAIDClassifier(CHAIDEstimator, ClassifierMixin):

	def __init__(self, config = {}):
		super(CHAIDClassifier, self).__init__(config = config)

	def _target_type(self):
		return "categorical"

	def fit(self, X, y, **fit_params):
		classes, y_encoded = numpy.unique(y, return_inverse = True)
		self.classes_ = classes
		y_encoded = Series(y_encoded, name = y.name)
		super(CHAIDClassifier, self).fit(X, y_encoded, **fit_params)
		return self

	def predict(self, X):
		raise NotImplementedError()

	def predict_proba(self, X):
		raise NotImplementedError()

class CHAIDRegressor(CHAIDEstimator, RegressorMixin):

	def __init__(self, config = {}):
		super(CHAIDRegressor, self).__init__(config = config)

	def _target_type(self):
		return "continuous"

	def predict(self, X):
		raise NotImplementedError()


from pandas import Categorical, DataFrame, Index, Series
from sklearn.base import clone, BaseEstimator, TransformerMixin

import inspect
import numpy
import pandas
import sys
import types

try:
	# Pandas 2.X
	iso8601_format = "ISO8601"

	pandas.to_datetime("2023-12-03", format = iso8601_format)
except ValueError:
	# Pandas 1.X
	iso8601_format = "%Y-%m-%dT%H:%M:%S.%f"

def cast(X, dtype):
	if isinstance(dtype, str) and dtype.startswith("datetime64"):
		func = lambda x: to_pydatetime(x, dtype)
		return dt_transform(X, func)
	else:
		if not hasattr(X, "astype"):
			X = numpy.asarray(X)
		return X.astype(dtype)

def common_dtype(X):
	if hasattr(X, "dtype"):
		return X.dtype
	elif hasattr(X, "dtypes"):
		dtypes = set(X.dtypes)
		if len(dtypes) != 1:
			raise ValueError(dtypes)
		return next(iter(dtypes))
	else:
		raise ValueError()

def to_numpy(X):
	if hasattr(X, "to_numpy"):
		return X.to_numpy()
	return X

def is_1d(X):
	shape = X.shape
	if len(shape) == 1:
		return True
	elif (len(shape) == 2 and shape[1] == 1):
		return True
	else:
		return False

def to_1d(X):
	if isinstance(X, (Categorical, Series)):
		return X
	elif isinstance(X, DataFrame):
		columns = X.columns
		if len(columns) == 1:
			return X[columns[0]]
	X = numpy.asarray(X)
	shape = X.shape
	if len(shape) == 1:
		return X
	elif (len(shape) == 2) and (shape[1] == 1):
		return X.ravel()
	else:
		raise ValueError("Expected 1d array or 2d column vector array, got {0}d array of shape {1}".format(len(shape), str(shape)))

def dt_transform(X, func):
	if hasattr(X, "applymap"):
		return X.applymap(func)
	shape = X.shape
	if len(shape) > 1:
		X = X.ravel()
	Xt = func(X)
	Xt = to_numpy(Xt)
	if len(shape) > 1:
		Xt = Xt.reshape(shape)
	return Xt

def to_pydatetime(X, dtype):
	Xt = pandas.to_datetime(X, yearfirst = True, format = iso8601_format, origin = "unix")
	if hasattr(Xt, "dt"):
		Xt = Xt.dt
	if dtype == "datetime64[D]":
		Xt = Xt.floor("D")
	elif dtype == "datetime64[s]":
		Xt = Xt.floor("S")
	else:
		raise ValueError(dtype)
	return Xt.to_pydatetime()

def ensure_def(expr, env):
	expr_code = compile(expr, "<string>", "exec")

	name = expr_code.co_names[0]

	if name in globals():
		return globals()[name]

	try:
		return env[name]
	except KeyError:
		eval(expr_code, env)

		return env[name]

class Evaluatable:

	def __init__(self, expr, function_defs = []):
		if isinstance(expr, str):
			if "\n" in expr:
				raise ValueError()
			self.expr = expr
		else:
			raise TypeError()

		def to_source(function_def):
			if isinstance(function_def, str):
				if "\n" not in function_def:
					raise ValueError()
				if "def" not in function_def:
					raise ValueError()
				return function_def
			elif isinstance(function_def, types.FunctionType):
				return inspect.getsource(function_def)
			else:
				raise TypeError()

		self.function_defs = [to_source(function_def) for function_def in function_defs]

	def setup(self, env):
		for function_def in self.function_defs:
			ensure_def(function_def, env)
		main_function_def = "def _evaluate(X):\n\treturn ({})".format(self.expr)
		ensure_def(main_function_def, env)

	def evaluate(self, X, env):
		func = env["_evaluate"]
		return func(X)

class Expression(Evaluatable):

	def __init__(self, expr, function_defs = []):
		super(Expression, self).__init__(expr = expr, function_defs = function_defs)

class Predicate(Evaluatable):

	def __init__(self, expr, function_defs = []):
		super(Predicate, self).__init__(expr = expr, function_defs = function_defs)

def to_expr(expr):
	if isinstance(expr, str):
		return expr
	elif isinstance(expr, types.FunctionType):
		if expr.__code__.co_argcount != 1:
			raise ValueError()
		if expr.__code__.co_varnames[0] != "X":
			raise ValueError()
		return inspect.getsource(expr)
	elif isinstance(expr, Evaluatable):
		return expr
	else:
		raise TypeError()

def to_expr_func(expr, modules = ["math", "numpy", "pandas"]):
	env = dict()

	# Import modules
	for module in modules:
		exec("import {}".format(module), env)

	if isinstance(expr, str):
		if "\n" not in expr:

			def evaluate(x):
				env["X"] = x
				return eval(expr, env)

			return evaluate
		else:
			func = ensure_def(expr, env)
			return lambda x: func(x)
	elif isinstance(expr, Evaluatable):
		expr.setup(env = env)
		return lambda x: expr.evaluate(x, env)
	else:
		raise TypeError()

def eval_rows(X, func, to_numpy = False, shape = None, dtype = None):
	if hasattr(X, "apply"):
		if isinstance(X, Series):
			Xt = X.apply(func)
		else:
			Xt = X.apply(func, axis = 1)
		if dtype is not None:
			Xt = Xt.astype(dtype)
		if to_numpy:
			Xt = Xt.to_numpy()
			if shape is not None:
				Xt = Xt.reshape(shape)
	else:
		nrow = X.shape[0]
		Xt = numpy.empty(shape = (nrow, ), dtype = (dtype if dtype is not None else object))
		for i in range(0, nrow):
			Xt[i] = func(X[i])
		if shape is not None:
			Xt = Xt.reshape(shape)
	return Xt

def fqn(obj):
	clazz = obj if inspect.isclass(obj) else obj.__class__
	return ".".join([clazz.__module__, clazz.__name__])

def is_instance_attr(obj, name):
	if not hasattr(obj, name):
		return False
	if name.startswith("__") and name.endswith("__"):
		return False
	v = getattr(obj, name)
	if isinstance(v, (types.BuiltinFunctionType, types.BuiltinMethodType, types.FunctionType, types.MethodType)):
		return False
	# See https://stackoverflow.com/a/17735709/
	attr_type = getattr(type(obj), name, None)
	if isinstance(attr_type, property):
		return False
	return True

def get_instance_attrs(obj):
	names = dir(obj)
	names = [name for name in names if is_instance_attr(obj, name)]
	return names

def sizeof(obj, with_overhead = False):
	if with_overhead:
		return sys.getsizeof(obj)
	return obj.__sizeof__()

def deep_sizeof(obj, with_overhead = False, verbose = False):
	# Primitive type values
	if obj is None:
		return obj.__sizeof__()
	elif isinstance(obj, (int, float, str, bool, numpy.int64, numpy.float32, numpy.float64)):
		return obj.__sizeof__()
	# Iterables
	elif isinstance(obj, list):
		sum = sizeof([], with_overhead = with_overhead) # Empty list
		for v in obj:
			v_sizeof = deep_sizeof(v, with_overhead = with_overhead, verbose = False)
			sum += v_sizeof
		return sum
	elif isinstance(obj, tuple):
		sum = sizeof((), with_overhead = with_overhead) # Empty tuple
		for i, v in enumerate(obj):
			v_sizeof = deep_sizeof(v, with_overhead = with_overhead, verbose = False)
			sum += v_sizeof
		return sum
	# Numpy ndarrays
	elif isinstance(obj, numpy.ndarray):
		sum = sizeof(obj, with_overhead = with_overhead) # Array header
		sum += (obj.size * obj.itemsize) # Array content
		return sum
	# Reference type values
	else:
		qualname = fqn(obj)
		# Restrict the circle of competence to Scikit-Learn classes
		if not (qualname.startswith("_abc.") or qualname.startswith("sklearn.")):
			raise TypeError("The object (class {0}) is not supported ".format(qualname))
		sum = sizeof(object(), with_overhead = with_overhead) # Empty object
		names = get_instance_attrs(obj)
		if names:
			if verbose:
				print("| Attribute | `type(v)` | `deep_sizeof(v)` |")
				print("|---|---|---|")
			for name in names:
				v = getattr(obj, name)
				v_type = type(v)
				v_sizeof = deep_sizeof(v, with_overhead = with_overhead, verbose = False)
				sum += v_sizeof
				if verbose:
					print("| {} | {} | {} |".format(name, v_type, v_sizeof))
		return sum

class Reshaper(BaseEstimator, TransformerMixin):

	def __init__(self, newshape):
		self.newshape = newshape

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		return X.reshape(self.newshape)

class Slicer(BaseEstimator, TransformerMixin):

	def __init__(self, start = None, stop = None, step = None):
		self.start = start
		self.stop = stop
		if (step is not None) and (step <= 0):
			raise ValueError("Step must be positive integer")
		self.step = step

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		rows = slice(None, None)
		columns = slice(self.start, self.stop, self.step)
		if isinstance(X, DataFrame):
			return X.iloc[rows, columns]
		else:
			return X[rows, columns]



from pandas import DataFrame

class FeatureMap(DataFrame):

	def __init__(self, data):
		super(FeatureMap, self).__init__(data = data, columns = ["id", "name", "type"])

	def save(self, path):
		self.to_csv(path, sep = "\t", header = False, index = False)

def make_feature_map(df, enable_categorical = True, category_to_indicator = True):
	entries = []

	feature_names = df.columns.tolist()

	for feature_name in feature_names:
		dtype = df[feature_name].dtype

		if dtype == "category" and enable_categorical:
			if category_to_indicator:
				categories = dtype.categories
				for category in categories:
					entries.append(("{}={}".format(feature_name, category), "i"))
			else:
				entries.append((feature_name, "c"))
		elif dtype in ("int8", "int16", "int32", "int64", "uint8", "uint16", "uint32", "uint64", "Int8", "Int16", "Int32", "Int64"):
			entries.append((feature_name, "int"))
		elif dtype in ("float16", "float32", "float64", "Float32", "Float64"):
			entries.append((feature_name, "q"))
		elif dtype in ("bool", "boolean"):
			entries.append((feature_name, "i"))
		else:
			raise ValueError(dtype)

	entries = [(idx, entry[0], entry[1]) for idx, entry in enumerate(entries)]

	return FeatureMap(entries)